@@ -11,7 +11,9 @@
  **********************************************************************/
 
 #include <stdio.h>
+#include <string.h>
 #include <unistd.h>
+#include <ctype.h>
 #include <sys/types.h>
 #include <sys/stat.h>
 #include <fcntl.h>
@@ -27,6 +29,44 @@
  *
  ********************************************/
 
+int state_validate_str(const char *str)
+{
+	const int len = strlen(str);
+	int i;
+	/* Spaces are ok, \n and \r not.
+	 * alpha, digits, +, -, @ are ok
+	 * These characters must be LaTeX safe, so no {}
+	 * Also they can be passed to some external script,
+	 * so they must obey restrictions.
+	 */
+	for (i=0; i<len; i++) {
+		if (isalnum(str[i]))
+			continue;
+		if (str[i] == ' ' || str[i] == '+' || str[i] == '@' || 
+		    str[i] == '-' || str[i] == '.' || str[i] == ',' ||
+		    str[i] == '*')
+			continue;
+		return 0; /* False */
+	}
+	return 1;
+}
+
+static char *_strtok(char *input, const char *delim)
+{
+	static char *position;
+	if (input != NULL)
+		position = input;
+	char *token = strsep(&position, delim); /* Non C99 function */
+
+	/* Cut token at any \n found */
+	if (token) {
+		char *pos = strchr(token, '\n');
+		if (pos)
+			*pos = '\0';
+	}
+	return token;
+}
+
 /* Returns name to user state file */
 static char *_state_file(const char *username, const char *filename)
 {
@@ -138,7 +178,9 @@ int state_lock(state *s)
 	s->lockname = malloc(strlen(s->filename) + 4 + 1);
 	if (!s->lockname)
 		return 1;
-	assert(sprintf(s->lockname, "%s.lck", s->filename) > 0);
+
+	ret = sprintf(s->lockname, "%s.lck", s->filename);
+	assert(ret > 0);
 
 	/* Open/create lock file */
 	fd = open(s->lockname, O_WRONLY|O_CREAT, S_IWUSR|S_IRUSR);
@@ -228,7 +270,7 @@ static int _rng_read(const char *device, const char *msg, unsigned char *buf, co
 	}
 
 	puts(
-		"Hint: Move your mouse, cause some disc activity\n"
+		"Hint: Move your mouse, cause some disc activity (`find /` is good)\n"
 		"or type on keyboard to make the progress faster.\n");
 
 	for (i=0; i<count; i++) {
@@ -247,20 +289,39 @@ static int _rng_read(const char *device, const char *msg, unsigned char *buf, co
 /**********************************************
  * Interface functions for managing state files
  **********************************************/
+
+static const int _version = 2;
+static const int _fields = 8;
+static const char *_delim = ":";
+
+enum {
+	FIELD_VERSION = 0,
+	FIELD_KEY,
+	FIELD_COUNTER,
+	FIELD_LATEST_CARD,
+	FIELD_CODE_LENGTH,
+	FIELD_FLAGS,
+	FIELD_LABEL,
+	FIELD_CONTACT,
+};
+
 int state_load(state *s)
 {
+	/* State file should never be larger than 160 bytes */
+	char buff[512];
+	/* How many bytes are in buff? */
+	unsigned int buff_size = 0; 
+
 	/* Did we lock it here? */
 	int locked = 0;
 
+	char *field[_fields]; /* Fields in file */
+
+	int i;
 	int ret = 0;
-	FILE *f;
+	FILE *f = NULL;
 
-	/* TODO: Maybe rewrite using strtok? */
-
-	if (s->filename == NULL) {
-		print(PRINT_CRITICAL, "State data not initialized?\n");
-		return 1;
-	}
+	assert(s->filename != NULL);
 
 	if (_state_file_permissions(s) != 0) {
 		print(PRINT_NOTICE,
@@ -281,114 +342,117 @@ int state_load(state *s)
 
 	f = fopen(s->filename, "r");
 	if (!f) {
-		print_perror(PRINT_ERROR, "Unable to open %s for reading. Have you tried -k option?", 
-			     s->filename);
-		goto error;
-	}
-
-	ret = mpz_inp_str(s->sequence_key, f, STATE_BASE);
-	if (ret == 0) {
-		print_perror(PRINT_ERROR, 
-			     "Error while reading sequence key from %s", 
-			     s->filename);
-		goto error;
-	}
-
-	ret = mpz_inp_str(s->counter, f, STATE_BASE);
-	if (ret == 0) {
-		print_perror(PRINT_ERROR, 
-			     "Error while reading counter from %s",
-			     s->filename);
-		goto error;
-	}
-
-	ret = mpz_inp_str(s->furthest_printed, f, STATE_BASE);
-	if (ret == 0) {
 		print_perror(PRINT_ERROR,
-			     "Error while reading number of "
-			     "last printed passcode from %s", s->filename);
-		goto error;
-	}
-
-	ret = fscanf(f, "%u\n", &s->code_length);
-	if (ret != 1) {
-		print_perror(PRINT_ERROR, 
-			     "Error while reading passcode length from %s",
+			     "Unable to open %s for reading.",
 			     s->filename);
 		goto error;
 	}
 
-	ret = fscanf(f, "%u", &s->flags);
-	if (ret != 1) {
-		print_perror(PRINT_ERROR, "Error while reading flags from %s",
-			     s->filename);
+	/* Read all file into buffer */
+	buff_size = fread(buff, 1, sizeof(buff), f);
+	if (buff_size < 10) {
+		/* This can't hold correct state */
+		print(PRINT_NOTICE, 
+		      "State file %s is invalid\n", s->filename);
 		goto error;
 	}
 
-	/* Read a whitecharacter after flags */
-	if (fgetc(f) != '\n') {
-		print_perror(PRINT_ERROR, "Syntax error in %s.", s->filename);
+	/* Split into fields */
+	for (i=0; i<_fields; i++) {
+		field[i] = _strtok(i == 0 ? buff : NULL, _delim);
+		if (field[i] == NULL) {
+			print(PRINT_ERROR, "State file invalid. Not enough fields.\n");
+			goto error;
+		}
+	}
+
+	if (_strtok(NULL, _delim) != NULL) {
+		print(PRINT_ERROR, "State file invalid. Too much fields.\n");
 		goto error;
 	}
 
-	if (fgets(s->label, sizeof(s->label), f) == NULL) {
-		/* Nothing read, there should be at least one \n */
-		print_perror(PRINT_ERROR, "Error while reading label from %s"
-			     ", unexpected end of file",
-			     s->filename);
+	/* Parse fields */
+	if (sscanf(field[FIELD_VERSION], "%u", &ret) != 1) {
+		print(PRINT_ERROR, "Error while parsing state file version.\n");
 		goto error;
 	}
 
-	if (s->label[strlen(s->label) - 1] != '\n') {
-		/* \n is put there if we find end of file,
-		 * it's lack might be caused by too long entry
-		 * at end of file 
-		 */
-		print_perror(PRINT_ERROR, "Garbage near label data in %s.", s->filename);
+	if (ret != _version) {
+		print(PRINT_ERROR, "State file version is incompatible. Recreate key.\n");
+		goto error;		
+	}
+
+	if (mpz_set_str(s->sequence_key, field[FIELD_KEY], 62) != 0) {
+		print(PRINT_ERROR, "Error while parsing sequence key.\n");
 		goto error;
 	}
 
-	if (fgets(s->contact, sizeof(s->contact), f) == NULL) {
-		/* Nothing read, there should be at least one \n */
-		print_perror(PRINT_ERROR, "Error while reading contact data from %s"
-			     ", unexpected end of file",
-			     s->filename);
+	if (mpz_set_str(s->counter, field[FIELD_COUNTER], 62) != 0) {
+		print(PRINT_ERROR, "Error while parsing counter.\n");
 		goto error;
 	}
 
-	if (s->contact[strlen(s->contact) - 1] != '\n') {
-		/* \n is put there if we find end of file,
-		 * it's lack might be caused by too long entry
-		 * at end of file 
-		 */
-		print_perror(PRINT_ERROR, "Garbage near contact data at end of %s.", s->filename);
+	if (mpz_set_str(s->furthest_printed, 
+			field[FIELD_LATEST_CARD], 62) != 0) {	
+		print(PRINT_ERROR,
+		      "Error while parsing number "
+		      "of latest printed passcard\n");
 		goto error;
 	}
 
-	/* Strip \n from the end of label and contact */
-	s->label[strlen(s->label) - 1] = '\0';
-	s->contact[strlen(s->contact) - 1] = '\0';
+	if (sscanf(field[FIELD_CODE_LENGTH], "%u", &s->code_length) != 1) {
+		print(PRINT_ERROR, "Error while parsing passcode length\n");
+		goto error;
+	}
+
+	if (sscanf(field[FIELD_FLAGS], "%u", &s->flags) != 1) {
+		print(PRINT_ERROR, "Error while parsing flags\n");
+		goto error;
+	}
+
+	/* Copy label and contact */
+	strncpy(s->label, field[FIELD_LABEL], sizeof(s->label)-1);
+	strncpy(s->contact, field[FIELD_CONTACT], sizeof(s->contact)-1);
+
+	if (s->label[sizeof(s->label)-1] != '\0') {
+		print(PRINT_ERROR, "Label field too long\n");
+		goto error;
+	}
+
+	if (s->contact[sizeof(s->contact)-1] != '\0') {
+		print(PRINT_ERROR, "Contact field too long\n");
+		goto error;
+	}
+
+	if (!state_validate_str(s->label)) {
+		print(PRINT_ERROR, "Illegal characters in label\n");
+		goto error;
+	}
+
+	if (!state_validate_str(s->contact)) {
+		print(PRINT_ERROR, "Illegal characters in contact\n");
+		goto error;
+	}
 
 	/* Everything is read. Now - check if it's correct */
-	/* TODO, FIXME */
 	if (mpz_sgn(s->sequence_key) == -1) {
 		print(PRINT_ERROR, 
-		      "Read a negative sequence key. File %s is corrupted.\n",
-		      s->filename);
+		      "Read a negative sequence key. "
+		      "State file is invalid\n");
 		goto error;
 	}
 
 	if (mpz_sgn(s->counter) == -1) {
 		print(PRINT_ERROR, 
-		      "Read a negative counter. File %s is corrupted.\n",
-		      s->filename);
+		      "Read a negative counter. "
+		      "State file is corrupted.\n");
 		goto error;
 	}
 
 	if (mpz_sgn(s->furthest_printed) == -1) {
 		print(PRINT_ERROR, 
-		      "Last printed card has a negative counter. File %s is corrupted.\n",
-		      s->filename);
+		      "Latest printed card is negative. "
+		      "State file is corrupted.\n");
 		goto error;
 	}
 
@@ -412,6 +476,7 @@ int state_load(state *s)
 	return 0;
 
 error:
+	memset(buff, 0, sizeof(buff));
 	if (locked && state_unlock(s) != 0) {
 		print(PRINT_ERROR, "Error while unlocking state file!\n");
 	}
@@ -421,16 +486,23 @@ error:
 
 int state_store(state *s)
 {
-	/* Rewrite maybe using gmp_scanf? */
-	int ret;
+	/* Return value, by default return error */
+	int ret = 1;
+
+	/* State file */
 	FILE *f;
 
+	/* Did we lock the file? */
 	int locked = 0;
 
-	if (s->filename == NULL) {
-		print(PRINT_CRITICAL, "State data not initialized?\n");
-		return 1;
-	}
+	/* Converted state parts */
+	char *sequence_key = NULL;
+	char *counter = NULL;
+	char *latest_card = NULL;
+
+	int tmp;
+
+	assert(s->filename != NULL);
 
 	if (s->lock_fd <= 0) {
 		print(PRINT_NOTICE, 
@@ -454,50 +526,31 @@ int state_store(state *s)
 	}
 
 	/* Write using ascii safe approach */
-	ret = mpz_out_str(f, STATE_BASE, s->sequence_key);
-	if (ret == 0) {
-		print(PRINT_ERROR, "Error while saving sequence key\n");
+	sequence_key = mpz_get_str(NULL, STATE_BASE, s->sequence_key);
+	counter = mpz_get_str(NULL, STATE_BASE, s->counter);
+	latest_card = mpz_get_str(NULL, STATE_BASE, s->furthest_printed);
+
+	if (!sequence_key || !counter || !latest_card) {
+		print(PRINT_ERROR, "Error while converting numbers\n");
 		goto error;
 	}
 
-	if (fputc('\n', f) == EOF) {
-		print(PRINT_ERROR, "Error while writting to %s\n", 
-		      s->filename);
+	const char d = _delim[0];
+	tmp = fprintf(f, "%d%c%s%c%s%c%s%c%u%c%u%c%s%c%s\n",
+		      _version, d,
+		      sequence_key, d, counter, d, latest_card, d,
+		      s->code_length, d, s->flags, d, s->label, d, s->contact);
+	if (tmp <= 0) {
+		print(PRINT_ERROR, "Error while writing data to state file.");
 		goto error;
 	}
-
-	ret = mpz_out_str(f, STATE_BASE, s->counter);
-	if (ret == 0) {
-		print(PRINT_ERROR, "Error while saving counter\n");
-		goto error;
-	}
-
-	if (fputc('\n', f) == EOF) {
-		print(PRINT_ERROR, "Error while writting to %s\n", 
-		      s->filename);
-		goto error;
-	}
-
-	ret = mpz_out_str(f, STATE_BASE, s->furthest_printed);
-	if (ret == 0) {
-		print(PRINT_ERROR, 
-		      "Error while saving number of last printed passcode\n");
-		goto error;
-	}
-
-	ret = fprintf(f, "\n%u\n%u\n%s\n%s\n", s->code_length, s->flags, s->label, s->contact);
-	if (ret <= 0) {
-		print(PRINT_ERROR, 
-		      "Error while writting passlength, flags, label and contact data to %s\n",
-		      s->filename);
-		goto error;
-	}
-
-	/* lock? */
-	print(PRINT_NOTICE, "State file written\n");
-
-	fflush(f);
-	fclose(f);
+	
+	tmp = fflush(f);
+	tmp += fclose(f);
+	if (tmp != 0)
+		print_perror(PRINT_ERROR, "Error while flushing/closing state file");
+	else
+		print(PRINT_NOTICE, "State file written\n");
 
 	/* It might fail, but shouldn't
 	 * Also we just want to ensure others 
@@ -508,18 +561,17 @@ int state_store(state *s)
 		      "Key might be world-readable!\n");
 	}
 
-	if (locked && state_unlock(s) != 0) {
-		print(PRINT_ERROR, "Error while unlocking state file!\n");
-	}
-	return 0;
+	ret = 0; /* More less fine */
 
 error:
 	if (locked && state_unlock(s) != 0) {
 		print(PRINT_ERROR, "Error while unlocking state file!\n");
 	}
 
-	fclose(f);
-	return 1;
+	free(sequence_key);
+	free(counter);
+	free(latest_card);
+	return ret;
 }
 
 /******************************************
@@ -532,6 +584,9 @@ int state_init(state *s, const char *username, const char *configfile)
 	const char code_mask[] =
 		"000000000000000000000000FFFFFFFF";
 
+	int ret;
+
+
 	assert(sizeof(salt_mask) == 33);
 
 	mpz_init(s->counter);
@@ -540,8 +595,11 @@ int state_init(state *s, const char *username, const char *configfile)
 	mpz_init(s->current_card);
 	mpz_init(s->max_card);
 	mpz_init(s->max_code);
-	assert(mpz_init_set_str(s->salt_mask, salt_mask, 16) == 0);
-	assert(mpz_init_set_str(s->code_mask, code_mask, 16) == 0);
+
+	ret = mpz_init_set_str(s->salt_mask, salt_mask, 16);
+	assert(ret == 0);
+	ret = mpz_init_set_str(s->code_mask, code_mask, 16);
+	assert(ret == 0);
 
 	s->code_length = 4;
 	s->flags = FLAG_SHOW;
@@ -593,9 +651,9 @@ void state_fini(state *s)
 
 int state_key_generate(state *s, const int salt)
 {
-	unsigned char entropy_pool[128]; /* 1024 bits */
+	unsigned char entropy_pool[64]; /* 512 bits */
 	unsigned char key_bin[32];
-	/* TODO: implement salting */
+
 	print(PRINT_NOTICE, "Generating new %s key.\n", salt ? "salted" : "not salted");
 
 	/* Gather entropy from random, then fallback to urandom... */
